import 'dart:async';
import 'dart:convert';
import 'storage/html.dart' if (dart.library.io) 'storage/io.dart';

class GetStorage {
  static final Map<String, GetStorage> _sync = {};

  StreamSubscription<Map<String, dynamic>> listen(
          void onData(Map<String, dynamic> event),
          {Function onError,
          void onDone(),
          bool cancelOnError}) =>
      stream.listen(onData,
          onError: onError, onDone: onDone, cancelOnError: cancelOnError);

  static Future<bool> init([String container = "GetStorage"]) {
    if (container == null) {
      throw 'key can not be null';
    }
    return GetStorage(container).initStorage;
  }

  factory GetStorage(
      [String container = "GetStorage",
      String path,
      Map<String, dynamic> initialData]) {
    if (_sync.containsKey(container)) {
      return _sync[container];
    } else {
      final instance = GetStorage._internal(container, path, initialData);
      _sync[container] = instance;
      return instance;
    }
  }

  void dispose() {
    _concrete?.dispose();
  }

  GetStorage._internal(String key,
      [String path, Map<String, dynamic> initialData]) {
    _concrete = StorageImpl(key, path);
    _initialData = initialData;

    initStorage = Future<bool>(() async {
      await this._init();
      return true;
    });
  }

  Future<void> _init() async {
    try {
      await _concrete.init(_initialData);
    } on Error catch (err) {
      throw err;
    }
  }

  T read<T>(String key) {
    return _concrete.read(key);
  }

  Future<void> write(String key, value, [EncodeObject objectToEncode]) async {
    final _encoded =
        json.encode(objectToEncode != null ? objectToEncode(value) : value);
    await _concrete.write(key, json.decode(_encoded));

    return _tryFlush();
  }

  Future<void> remove(String key) async {
    await _concrete.remove(key);
    return _tryFlush();
  }

  Future<void> clear() async {
    await _concrete.clear();
    return _tryFlush();
  }

  Future<void> _tryFlush() async {
    if (_lockDatabase != null) {
      await _lockDatabase;
    }
    _lockDatabase = _flush();
    return _lockDatabase;
  }

  Future<void> _flush() async {
    try {
      await _concrete.flush();
    } catch (e) {
      rethrow;
    }
    return;
  }

  StorageImpl _concrete;

  Future<bool> initStorage;

  Future<void> _lockDatabase;

  Stream<Map<String, dynamic>> get stream => _concrete.stream;

  Map<String, dynamic> _initialData;
}

typedef EncodeObject = Object Function(Object);
